import {ARBatch} from "./ARBatch";
import {PLCRow} from "kaneoh-data-types/types/dtos/PLCDto";
import {ApprovalRequestDifferences} from "kaneoh-data-types/types/records/ApprovalRequestRecord";
import {ApprovalRequestRuleCollection} from "kaneoh-data-types/types/records/PlatformConfigRecord";
import {UserDto} from "kaneoh-data-types/types/dtos/UserDto";
import * as _ from "underscore";
import {generateApprovalRequestDifferencesByRules} from "./generateApprovalRequestDifferencesByRules";
import {generateDifferences} from "./generateDifferences";
import {revertFields} from "./revertFields";

export class PLCARBatch implements ARBatch<PLCRow> {
    private rules: ApprovalRequestRuleCollection;
    private comparisons: Array<PLCRow>;
    private user: UserDto;

    constructor(rules: ApprovalRequestRuleCollection, comparisons: Array<PLCRow>, user: UserDto) {
        this.rules = rules;
        this.comparisons = comparisons;
        this.user = user;
    }

    public shouldBeApproval(dtos: Array<PLCRow>): boolean {
        return this.build(dtos).length > 0;
    }

    public revert(dtos: Array<PLCRow>): Array<PLCRow> {
        const diffs = this.build(dtos);
        _(dtos).map((dto) => {
            const dtoDiffs = _(diffs).where({sourceId: dto._id});
            revertFields(dto, dtoDiffs);
            dto.children = this.revert(dto.children);
        });
        return dtos;
    }

    public build(dtos: Array<PLCRow>): ApprovalRequestDifferences {
        const user = this.findOwnerRules(this.rules, this.user);
        if (user.length > 0) {
            const differences = this.diffs(dtos);
            return generateApprovalRequestDifferencesByRules(this.rules, differences);
        }
        return [];
    }

    public updateComparison(comparisons: Array<PLCRow>): void {
        this.comparisons = comparisons;
    }

    private findOwnerRules(rules: ApprovalRequestRuleCollection, owner: UserDto): ApprovalRequestRuleCollection {
        return _(rules)
            .chain()
            .filter((it) => {
                if (it.byRole === "*") {
                    return true;
                }
                return _(owner.teams).contains(it.byRole) || _(it.byUserIds).contains(owner._id);
            })
            .value();
    }

    private diffs(dtos: Array<PLCRow>, comparisons: Array<PLCRow> = this.comparisons): ApprovalRequestDifferences {
        const comparisonIds = _(comparisons).pluck("_id");
        return _(dtos)
            .chain()
            .filter((dto) => {
                return _(comparisonIds).contains(dto._id);
            })
            .map((dto) => {
                const comparison = _(comparisons).findWhere({_id: dto._id});
                return this.diff(dto, comparison);
            })
            .flatten()
            .value();
    }

    private diff(dto: PLCRow, comparison: PLCRow): ApprovalRequestDifferences {
        const diffs = generateDifferences<PLCRow>(comparison, dto);
        const childrenDiffs = this.diffs(dto.children, comparison.children);
        return _.union(diffs, childrenDiffs);
    }
}
